class BaseClient:
    def get(self, key: str):
        """Get the value of ``key``

        Args:
            key (``str``):
                The key to get.
        """
        raise NotImplementedError

    def set(self, key: str, value):
        """Get the value of ``key``

        Args:
            key (``str``):
                The key.

            value (``Any``):
                The value to set for ``key``.

        Returns:
            :py:class:`bool`: ``True`` on success.
        """
        raise NotImplementedError

    def delete(self, key: str):
        """Delete ``key`` from database

        Args:
            key (``str``):
                The key to delete.
        """
        raise NotImplementedError

    def commit(self):
        """Commit the current changes

        Returns:
            :py:class:`bool`: ``True`` on success.
        """
        raise NotImplementedError

    def exists(self, key: str):
        """Check if ``key`` already exists in database

        Args:
            key (``str``):
                The key to search for.

        Returns:
            :py:class:`bool`: ``True`` if found, otherwise ``False``.
        """
        raise NotImplementedError

    def rename(self, key: str, new_key: str):
        """Rename ``key`` with ``new_key``

        Args:
            key (``str``):
                The key to rename.

            new_key (``str``):
                The key to rename with.

        Returns:
            :py:class:`bool`: ``True`` if renamed, otherwise ``False``.
        """
        raise NotImplementedError

    def keys(self, like: str = "%"):
        """Return list of keys in database with the given pattern

        Args:
            like (``str``, *optional*):
                SQLite LIKE operator. Defaults to ``%`` (all keys).

        Returns:
            :py:class:`list`:
                A list of :py:class:`Tuple` contains keys.

            :py:class:`None`:
                If there is no keys to return.
        """
        raise NotImplementedError

    def flush(self):
        """Flush and remove everything from the current database

        Returns:
            :py:class:`bool`: ``True`` on success.
        """
        raise NotImplementedError

    def close(self, optimize_database: bool = True):
        """Close database connection

        Args:
            optimize_database (``bool``, **optional**):
                Whether optimize database before closing or not. Defaults to ``True``.

        Returns:
            :py:class:`bool`: ``True`` on success.
        """
        raise NotImplementedError



import sqlite3
import logging

from concurrent.futures import ThreadPoolExecutor
from threading import Lock
from sys import version_info

logger = logging.getLogger(__name__)


class REQUEST:
    GET = "GET"
    SET = "SET"
    DELETE = "DELETE"
    COMMIT = "COMMIT"
    EXISTS = "EXISTS"
    RENAME = "RENAME"
    KEYS = "KEYS"
    FLUSH_DB = "FLUSH_DB"
    CLOSE = "CLOSE"


TABLE_STATEMENT = (
    'CREATE TABLE IF NOT EXISTS "{}" (k VARCHAR(4096) PRIMARY KEY, v BLOB)'
)


class Sqlite:
    def __init__(
        self,
        database: str,
        table_name: str,
        autocommit: bool,
        journal_mode: str,
        synchronous: str,
        encoder,
        workers: int,
    ) -> None:
        assert isinstance(database, str), "database must be str"
        assert isinstance(table_name, str), "table_name must be str"
        assert isinstance(autocommit, bool), "autocommit must be bool"
        assert isinstance(journal_mode, str), "journal_mode must be str"
        assert isinstance(synchronous, str), "synchronous must be str"

        self.database = database
        self.table_name = table_name
        self.autocommit = autocommit
        self.journal_mode = journal_mode
        self.synchronous = synchronous
        self.__encoder = encoder
        self.__workers = ThreadPoolExecutor(workers, "kvsqlite")
        self.__connection: sqlite3.Connection = self.__connect()
        self.__lock = Lock()

        self.is_running = True

    def request(self, request, key: str = None, value=None):
        return self.__workers.submit(self.procces_request, request, key, value)

    def procces_request(self, request, key: str = None, value=None):
        if not self.is_running:
            raise RuntimeError("Database is closed")

        logger.debug("Request={}, key={}".format(request, key))

        # TODO: Simplify request handlers

        if request == REQUEST.GET:
            try:
                query = self.__connection.execute(
                    'SELECT v FROM "{}" WHERE k = ?'.format(self.table_name),
                    (key,),
                ).fetchone()
                if query:
                    return self.__encoder.decode(query[0])
                else:
                    return None
            except Exception as e:
                logger.exception("SELECT statment error")
                raise e
        elif request == REQUEST.SET:
            with self.__lock:
                try:
                    query = self.__connection.execute(
                        'REPLACE INTO "{}" (k, v) VALUES(?,?)'.format(self.table_name),
                        (key, self.__encoder.encode(value)),
                    )
                    if query.rowcount > 0:
                        return True
                    else:
                        return False
                except Exception as e:
                    logger.exception("REPLACE INTO statment error")
                    raise e
        elif request == REQUEST.DELETE:
            with self.__lock:
                try:
                    query = self.__connection.execute(
                        'DELETE FROM "{}" WHERE k = ?'.format(self.table_name),
                        (key,),
                    )
                    if query.rowcount > 0:
                        return True
                    else:
                        return False
                except Exception as e:
                    logger.exception("DELETE statment error")
                    raise e
        elif request == REQUEST.COMMIT:
            try:
                self.__connection.commit()
                return True
            except Exception as e:
                logger.exception("COMMIT error")
                raise e
        elif request == REQUEST.EXISTS:
            try:
                query = self.__connection.execute(
                    'SELECT k FROM "{}" WHERE k = ?'.format(self.table_name),
                    (key,),
                ).fetchone()

                if query:
                    return True
                else:
                    return False
            except Exception as e:
                logger.exception("EXISTS error")
                raise e
        elif request == REQUEST.RENAME:
            try:
                query = self.__connection.execute(
                    'UPDATE OR IGNORE "{}" SET k = ? WHERE k = ?'.format(
                        self.table_name
                    ),
                    (value, key),
                )

                if query.rowcount > 0:
                    return True
                else:
                    return False
            except Exception as e:
                logger.exception("RENAME KEY error")
                raise e
        elif request == REQUEST.KEYS:
            try:
                query = self.__connection.execute(
                    'SELECT k FROM "{}" WHERE k LIKE ? ORDER BY rowid'.format(
                        self.table_name
                    ),
                    (value,),
                ).fetchall()
                if query:
                    return query
                else:
                    return None
            except Exception as e:
                logger.exception("SELECT KEYS statment error")
                raise e
        elif request == REQUEST.FLUSH_DB:
            with self.__lock:
                try:
                    self.__connection.execute('DROP TABLE "{}"'.format(self.table_name))
                    self.__connection.execute(TABLE_STATEMENT.format(self.table_name))
                    return True
                except Exception as e:
                    logger.exception("DROP TABLE statment error")
                    raise e
        elif request == REQUEST.CLOSE:
            with self.__lock:
                try:
                    if value:
                        self.__connection.execute("PRAGMA optimize")
                    self.__connection.close()
                    logger.info("Connection to {} closed".format(self.database))

                    if version_info.minor > 8:
                        self.__workers.shutdown(False, cancel_futures=True)
                    else:
                        self.__workers.shutdown(False)

                    self.is_running = False
                    return True
                except Exception as e:
                    logger.exception("On close error")
                    raise e

    def __connect(self):
        try:
            if self.autocommit:
                connection = sqlite3.connect(
                    self.database, isolation_level=None, check_same_thread=False
                )
            else:
                connection = sqlite3.connect(self.database, check_same_thread=False)

            # connection.row_factory = sqlite3.Row
            logger.info("Connected to {}".format(self.database))
        except Exception as e:
            logger.exception(
                "Error while opening sqlite3 self.__connection for database: {}".format(
                    self.database
                )
            )
            raise e

        try:
            connection.execute("PRAGMA journal_mode = {}".format(self.journal_mode))
            connection.execute("PRAGMA synchronous = {}".format(self.synchronous))
        except Exception as e:
            logger.exception("Error while executing PRAGMA statement")
            raise e

        try:
            connection.execute(TABLE_STATEMENT.format(self.table_name))
        except Exception as e:
            logger.exception("Error while executing CREATE TABLE statement")
            raise e

        return connection



import argparse
import asyncio
import random
import string
import kvsqlite
import time
import os
import psutil

parser = argparse.ArgumentParser(description="Benchmark kvsqlite")
parser.add_argument(
    "--query-count",
    type=int,
    help="Number of queryies to benchmark (Defaults to 100000)",
    default=100000,
)
parser.add_argument(
    "--db-path",
    type=str,
    help="Databse path (Defaults to benchmark_kvsqlite.sqlite)",
    default="benchmark_kvsqlite.sqlite",
)
args = parser.parse_args()

if args.query_count < 1:
    raise ValueError("--query-count must be greater than 1")


def random_string(length):
    return "".join(random.choice(string.ascii_letters) for _ in range(length))


PINK = "\033[95m"
GREEN = "\033[92m"
WARNING = "\033[93m"
ENDC = "\033[0m"


async def benchmark_set(db, keys):
    print(PINK, "================Benchmark set================", ENDC)
    timeing = 0
    start = time.perf_counter()
    start_memory = psutil.Process(os.getpid()).memory_info().rss
    print(
        WARNING,
        "-> Started with memory usage:",
        ENDC,
        start_memory,
    )
    for k, v in keys:
        latncey_start = time.perf_counter()
        await db.set(k, v)
        timeing += time.perf_counter() - latncey_start
    end_memory = psutil.Process(os.getpid()).memory_info().rss
    took = time.perf_counter() - start
    print(
        WARNING,
        "-> {} query took:{} {}".format(args.query_count, ENDC, took),
    )
    print(
        WARNING,
        "-> QRS:",
        ENDC,
        int(args.query_count / took),
    )
    print(
        WARNING,
        "-> Average latancey:",
        ENDC,
        timeing / args.query_count,
    )
    print(
        WARNING,
        "-> Current memory usage:{} {} ({}+{}{})".format(
            ENDC, end_memory, GREEN, (end_memory - start_memory), ENDC
        ),
        ENDC,
    )
    print(GREEN, "================Benchmark end================", ENDC)
    print()


async def benchmark_get(db, keys):
    print(PINK, "================Benchmark get================", ENDC)
    timeing = 0
    start = time.perf_counter()
    start_memory = psutil.Process(os.getpid()).memory_info().rss
    print(
        WARNING,
        "-> Started with memory usage:",
        ENDC,
        start_memory,
    )
    for k, v in keys:
        latncey_start = time.perf_counter()
        await db.get(k)
        timeing += time.perf_counter() - latncey_start
    end_memory = psutil.Process(os.getpid()).memory_info().rss
    took = time.perf_counter() - start
    print(
        WARNING,
        "-> {} query took:{} {}".format(args.query_count, ENDC, took),
    )
    print(
        WARNING,
        "-> QRS:",
        ENDC,
        int(args.query_count / took),
    )
    print(
        WARNING,
        "-> Average latancey:",
        ENDC,
        timeing / args.query_count,
    )
    print(
        WARNING,
        "-> Current memory usage:{} {} ({}+{}{})".format(
            ENDC, end_memory, GREEN, (end_memory - start_memory), ENDC
        ),
        ENDC,
    )
    print(GREEN, "================Benchmark end================", ENDC)
    print()


async def benchmark_exists(db, keys):
    print(PINK, "================Benchmark exists=============", ENDC)
    timeing = 0
    start = time.perf_counter()
    start_memory = psutil.Process(os.getpid()).memory_info().rss
    print(
        WARNING,
        "-> Started with memory usage:",
        ENDC,
        start_memory,
    )
    for k, v in keys:
        latncey_start = time.perf_counter()
        await db.exists(k)
        timeing += time.perf_counter() - latncey_start
    end_memory = psutil.Process(os.getpid()).memory_info().rss
    took = time.perf_counter() - start
    print(
        WARNING,
        "-> {} query took:{} {}".format(args.query_count, ENDC, took),
    )
    print(
        WARNING,
        "-> QRS:",
        ENDC,
        int(args.query_count / took),
    )
    print(
        WARNING,
        "-> Average latancey:",
        ENDC,
        timeing / args.query_count,
    )
    print(
        WARNING,
        "-> Current memory usage:{} {} ({}+{}{})".format(
            ENDC, end_memory, GREEN, (end_memory - start_memory), ENDC
        ),
        ENDC,
    )
    print(GREEN, "================Benchmark end================", ENDC)
    print()


async def benchmark_delete(db, keys):
    print(PINK, "================Benchmark delete=============", ENDC)
    timeing = 0
    start = time.perf_counter()
    start_memory = psutil.Process(os.getpid()).memory_info().rss
    print(
        WARNING,
        "-> Started with memory usage:",
        ENDC,
        start_memory,
    )
    for k, v in keys:
        latncey_start = time.perf_counter()
        await db.delete(k)
        timeing += time.perf_counter() - latncey_start
    end_memory = psutil.Process(os.getpid()).memory_info().rss
    took = time.perf_counter() - start
    print(
        WARNING,
        "-> {} query took:{} {}".format(args.query_count, ENDC, took),
    )
    print(
        WARNING,
        "-> QRS:",
        ENDC,
        int(args.query_count / took),
    )
    print(
        WARNING,
        "-> Average latancey:",
        ENDC,
        timeing / args.query_count,
    )
    print(
        WARNING,
        "-> Current memory usage:{} {} ({}+{}{})".format(
            ENDC, end_memory, GREEN, (end_memory - start_memory), ENDC
        ),
        ENDC,
    )
    print(GREEN, "================Benchmark end================", ENDC)
    print()


async def benchmark_concurrent_set(db, keys):
    print(PINK, "================Benchmark concurrent set=====", ENDC)

    futures = asyncio.gather(*[db.set(k, v) for k, v in keys])
    start = time.perf_counter()
    start_memory = psutil.Process(os.getpid()).memory_info().rss
    print(
        WARNING,
        "-> Started with memory usage:",
        ENDC,
        start_memory,
    )
    await futures
    end_memory = psutil.Process(os.getpid()).memory_info().rss
    took = time.perf_counter() - start
    print(
        WARNING,
        "-> {} query took:{} {}".format(args.query_count, ENDC, took),
    )
    print(
        WARNING,
        "-> QRS:",
        ENDC,
        int(args.query_count / took),
    )
    print(
        WARNING,
        "-> Current memory usage:{} {} ({}+{}{})".format(
            ENDC, end_memory, GREEN, (end_memory - start_memory), ENDC
        ),
        ENDC,
    )
    print(GREEN, "================Benchmark end================", ENDC)
    print()


async def benchmark_concurrent_get(db, keys):
    print(PINK, "================Benchmark concurrent get=====", ENDC)

    futures = asyncio.gather(*[db.get(k) for k, v in keys])
    start = time.perf_counter()
    start_memory = psutil.Process(os.getpid()).memory_info().rss
    print(
        WARNING,
        "-> Started with memory usage:",
        ENDC,
        start_memory,
    )
    await futures
    end_memory = psutil.Process(os.getpid()).memory_info().rss
    took = time.perf_counter() - start
    print(
        WARNING,
        "-> {} query took:{} {}".format(args.query_count, ENDC, took),
    )
    print(
        WARNING,
        "-> QRS:",
        ENDC,
        int(args.query_count / took),
    )
    print(
        WARNING,
        "-> Current memory usage:{} {} ({}+{}{})".format(
            ENDC, end_memory, GREEN, (end_memory - start_memory), ENDC
        ),
        ENDC,
    )
    print(GREEN, "================Benchmark end================", ENDC)
    print()


async def main():

    keys = []
    for _ in range(args.query_count):
        keys.append(
            (
                random_string(random.randint(1, 20)),
                random_string(random.randint(1, 20)),
            )
        )

    async with kvsqlite.Client(args.db_path) as db:
        print(WARNING, args.query_count, "Query benchmark", ENDC)
        await benchmark_set(db, keys)
        await benchmark_get(db, keys)
        await benchmark_exists(db, keys)
        await benchmark_delete(db, keys)

        await db.flush()

    async with kvsqlite.Client(args.db_path, workers=5) as db:
        print(
            WARNING,
            args.query_count,
            "Query benchmark with",
            db.workers,
            "workers",
            ENDC,
        )
        await benchmark_concurrent_set(db, keys)
        await benchmark_concurrent_get(db, keys)

        await db.flush()


asyncio.run(main())



__version__ = "0.1.2"
__copyright__ = "Copyright (c) 2023 AYMEN Mohammed ~ https://github.com/AYMENJD"
__license__ = "MIT License"

VERSION = __version__

__all__ = ["Client", "PickleEncoder", "StringEncoder", "sync"]

from .client import Client
from .encoders import PickleEncoder, StringEncoder
from . import sync



import logging

from typing import Any, List, Tuple, Union
from ..sqlite import Sqlite, REQUEST
from ..encoders import PickleEncoder
from ..base import BaseClient

logger = logging.getLogger(__name__)


class Client(BaseClient):
    def __init__(
        self,
        database: str,
        table_name: str = "kvsqlite",
        autocommit: bool = True,
        journal_mode: str = "WAL",
        synchronous: str = "NORMAL",
        default_encoder=PickleEncoder,
        workers: int = 2,
    ):
        """Kvsqlite synchronous client

        Args:
            database (``str``):
                Sqlite3 database path.

            table_name (``str``, *optional*):
                Table name to use, will be created if not exists. Defaults to ``kvsqlite``.

            autocommit (``bool``, *optional*):
                Whether `autocommit` database changes or not. Defaults to ``True``.

            journal_mode (``str``, *optional*):
                See https://www.sqlite.org/pragma.html#pragma_journal_mode. Defaults to ``WAL``.

            synchronous (``str``, *optional*):
                See https://www.sqlite.org/pragma.html#pragma_synchronous. Defaults to ``NORMAL``.

            default_encoder (``Callable``, *optional*):
                The encoder class which deal with the data sent/received by sqlite3. Defaults to :class:`kvsqlite.PickleEncoder`.

            workers (``int``, *optional*):
                The number of workers which process sqlite queries. Defaults to ``2``.
        """
        assert isinstance(database, str), "database must be str"
        assert isinstance(table_name, str), "table_name must be str"
        assert isinstance(autocommit, bool), "autocommit must be bool"
        assert isinstance(journal_mode, str), "journal_mode must be str"
        assert isinstance(synchronous, str), "synchronous must be str"
        assert callable(default_encoder), "default_encoder must be callable"
        assert isinstance(workers, int), "workers must be int"
        assert workers > 0, "workers must be greater than 0"

        assert hasattr(
            default_encoder, "encode"
        ), "{} must have an 'encode' function".format(default_encoder.__name__)
        assert hasattr(
            default_encoder, "decode"
        ), "{} must have an 'decode' function".format(default_encoder.__name__)

        self.database = database
        self.table_name = table_name
        self.autocommit = autocommit
        self.journal_mode = journal_mode
        self.synchronous = synchronous
        self.__encoder = default_encoder()
        self.workers = workers

        self.__sqlite = Sqlite(
            self.database,
            self.table_name,
            self.autocommit,
            self.journal_mode,
            self.synchronous,
            self.__encoder,
            self.workers,
        )

        logger.debug("Using {} as encoder".format(default_encoder.__name__))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.close()
        except Exception:
            logger.exception("Error on __exit__")

    def get(self, key: str) -> Any:
        assert isinstance(key, str), "key must be str"

        return self.__invoke(request=REQUEST.GET, key=key)

    def set(self, key: str, value) -> bool:
        assert isinstance(key, str), "key must be str"

        return self.__invoke(request=REQUEST.SET, key=key, value=value)

    def delete(self, key: str) -> bool:
        assert isinstance(key, str), "key must be str"

        return self.__invoke(request=REQUEST.DELETE, key=key)

    def commit(self) -> bool:
        return self.__invoke(request=REQUEST.COMMIT)

    def exists(self, key: str) -> bool:
        assert isinstance(key, str), "key must be str"

        return self.__invoke(request=REQUEST.EXISTS, key=key)

    def rename(self, key: str, new_key: str) -> bool:
        assert isinstance(key, str), "key must be str"
        assert isinstance(new_key, str), "new_key must be str"

        return self.__invoke(request=REQUEST.RENAME, key=key, value=new_key)

    def keys(self, like: str = "%") -> Union[List[Tuple[str]], None]:
        assert isinstance(like, str), "like must be str"

        return self.__invoke(request=REQUEST.KEYS, value=like)

    def flush(self) -> bool:
        return self.__invoke(request=REQUEST.FLUSH_DB)

    def close(self, optimize_database: bool = True) -> bool:
        assert isinstance(optimize_database, bool), "optimize_database must be bool"

        return self.__invoke(request=REQUEST.CLOSE, value=optimize_database)

    def __invoke(self, request, key=None, value=None):
        assert self.__sqlite.is_running, "Database is closed"

        future = self.__sqlite.request(request, key, value)
        return future.result()



from re import findall
from setuptools import setup, find_packages


with open("kvsqlite/__init__.py", "r") as f:
    version = findall(r"__version__ = \"(.+)\"", f.read())[0]

with open("README.md", "r") as f:
    readme = f.read()


setup(
    name="Kvsqlite",
    version=version,
    description="Easy-to-use synchronous/asynchronous key-value database backed by sqlite3.",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="AYMEN Mohammed",
    author_email="let.me.code.safe@gmail.com",
    url="https://github.com/AYMENJD/Kvsqlite",
    license="MIT",
    python_requires=">=3.8",
    project_urls={
        "Source": "https://github.com/AYMENJD/Kvsqlite",
        "Tracker": "https://github.com/AYMENJD/Kvsqlite/issues",
        "Documentation": "https://kvsqlite.rtfd.io/",
    },
    packages=find_packages(exclude=["docs"]),
    keywords=[
        "sync",
        "asyncio",
        "sqlite",
        "sqlite3",
        "key-value",
        "database",
        "redis",
    ],
)



from pickle import dumps, loads, HIGHEST_PROTOCOL
from sqlite3 import Binary


class PickleEncoder:
    """Encoder which uses pickle to serialize/deserialize the object"""

    def __init__(self) -> None:
        pass

    def encode(self, obj):
        return Binary(dumps(obj, protocol=HIGHEST_PROTOCOL))

    def decode(self, obj):
        return loads(bytes(obj))


class StringEncoder:
    """This encoder can be used instead of :class:`PickleEncoder`. This encoder accpets :py:class:`str` only"""

    def __init__(self) -> None:
        pass

    def encode(self, text):
        assert isinstance(text, str), "text is not str"
        return Binary(text.encode("utf-8"))

    def decode(self, text):
        assert isinstance(text, str), "text is not str"
        return text



import asyncio
import pytest
import random
import string
import kvsqlite

pytest_plugins = ("pytest_asyncio",)


def random_string(length):
    return "".join(random.choice(string.ascii_letters) for _ in range(length))


@pytest.mark.asyncio
async def test_key_value():
    keys = []
    values = []

    for _ in range(100):
        keys.append(random_string(random.randint(1, 20)))

    for _ in range(100):
        values.append(random_string(random.randint(1, 20)))

    async with kvsqlite.Client("test_kvsqlite.sqlite") as db:
        queries_n = 100000
        queries = []
        for _ in range(queries_n):
            op = random.randint(0, 2)
            key = random.choice(keys)
            value = random.choice(values)
            if op == 0:
                q = db.get(key)
            elif op == 1:
                q = db.delete(key)
            elif op == 2:
                q = db.set(key, value)
            queries.append((op, q))

        for op, query in queries:
            result = await query
            if op == 2:
                assert result
            elif op == 1:
                assert isinstance(result, bool)

        await db.flush()


@pytest.mark.asyncio
async def test_in_memory_key_value():
    keys = []
    values = []

    for _ in range(100):
        keys.append(random_string(random.randint(1, 20)))

    for _ in range(100):
        values.append(random_string(random.randint(1, 20)))

    async with kvsqlite.Client(":memory:") as db:
        queries_n = 100000
        queries = []
        for _ in range(queries_n):
            op = random.randint(0, 2)
            key = random.choice(keys)
            value = random.choice(values)
            if op == 0:
                q = db.get(key=key)
            elif op == 1:
                q = db.delete(key=key)
            elif op == 2:
                q = db.set(key=key, value=value)
            queries.append((op, q))

        for op, query in queries:
            result = await query
            if op == 2:
                assert result == True
            elif op == 1:
                assert isinstance(result, bool) == True

        await db.flush()


@pytest.mark.asyncio
async def test_concurrent_get():

    async with kvsqlite.Client("test_kvsqlite.sqlite", workers=5) as db:
        keys = []

        for _ in range(100):
            key = random_string(random.randint(1, 20))
            keys.append(key)
            await db.set(key=key, value=random_string(random.randint(1, 20)))

        futures = await asyncio.gather(
            *[db.get(random.choice(keys)) for _ in range(100000)]
        )

        for result in futures:
            assert isinstance(result, str) == True

        await db.flush()


@pytest.mark.asyncio
async def test_in_memory_concurrent_get():

    async with kvsqlite.Client(":memory:") as db:
        keys = []

        for _ in range(100):
            key = random_string(random.randint(1, 20))
            keys.append(key)
            await db.set(key=key, value=random_string(random.randint(1, 20)))

        futures = await asyncio.gather(
            *[db.get(random.choice(keys)) for _ in range(100000)]
        )

        for result in futures:
            assert isinstance(result, str) == True

        await db.flush()


@pytest.mark.asyncio
async def test_concurrent_set():

    async with kvsqlite.Client("test_kvsqlite.sqlite", workers=5) as db:

        futures = await asyncio.gather(
            *[
                db.set(
                    random_string(random.randint(1, 20)),
                    random_string(random.randint(1, 20)),
                )
                for _ in range(100000)
            ]
        )

        for result in futures:
            assert result == True

        await db.flush()


@pytest.mark.asyncio
async def test_in_memory_concurrent_set():

    async with kvsqlite.Client(":memory:") as db:

        futures = await asyncio.gather(
            *[
                db.set(
                    random_string(random.randint(1, 20)),
                    random_string(random.randint(1, 20)),
                )
                for _ in range(100000)
            ]
        )

        for result in futures:
            assert result == True

        await db.flush()


@pytest.mark.asyncio
async def test_key_exists():

    async with kvsqlite.Client("test_kvsqlite.sqlite") as db:
        keys = []

        for _ in range(10000):
            key = random_string(random.randint(1, 20))
            keys.append(key)
            assert (
                await db.set(key=key, value=random_string(random.randint(1, 20)))
                == True
            )

        for key in keys:
            result = await db.exists(key)
            assert result == True
